/**
  Raytracing demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2010-03-22 (Public Domain)
*/
#include <iostream>
#include <cmath>
#include <GL/glew.h> /* 
OpenGL Extensions Wrangler:  for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <GL/glut.h> /* OpenGL Utilities Toolkit, for GUI tools */
#include "ogl/glsl.h"
#include "ogl/glsl.cpp"
#include "ogl/minicam.h"
#include "osl/mat4.h"
#include "osl/mat4_inverse.h"
using osl::mat4;

#include <vector>

/** A bounding volume hierarchy, stored in a 2D texture */
class bvh_tex {
public:
	int w,h; /* size of 2D texture (in vec4 RGBA pixels) */
	std::vector<vec4> arr;
	vec4 root; // first BVH node to traverse
	int leaf_x; // column for leaf blocks
	int next_leaf; // row for next leaf block
	int next_node; // row for next node block
	
	bvh_tex(int w_,int h_) 
		:w(w_), h(h_), arr(w*h),
		root(0.0,0.0,-1.0,0.0),
		leaf_x(9), next_leaf(0), next_node(0)
		{}
	
	/* Access this pixel of the bvh texture */
	vec4 &at(const vec4 &tex) {
		int x=(int)tex.x, y=(int)tex.y;
		if (x<0 || x>=w || y<0 || y>=h) {
			printf("BVH: Out of bounds bvh access at (%d,%d), should be inside (%d,%d)\n",
				x,y,w,h);
			exit(1);
		}
		return arr[x+y*w];
	}
	
	/* Add a block of internal nodes in the bounding tree:
	     bounding volumes.  Render the given next link afterwards.  
	   Returns destination tex coords for the new block. */
	vec4 add_node_block(int n,const vec4 &nextlink) {
		float node_x=0.0; // x coordinate of start of node blocks
		vec4 tex(node_x,next_node++,n,node_x+n);
		at(tex+vec4(n,0.0,0.0,0.0))=nextlink; // add next link
		return tex;
	}
	
	/* Add another internal node at tex.  Increments tex.
	  Returns the nextlink pointer.  
	  Typical call pattern:
	     vec4 t=bvh.add_node_block(2,nextlink);
	     
	     vec4 ret_A=bvh.add_node_start(t,bound_A);
	     vec4 hit_A=bvh.add_...(...,ret_A);
	     bvh.add_node_end(t,hit_A);
	     
	     vec4 ret_B=bvh.add_node_start(t,bound_B);
	     vec4 hit_B=bvh.add_...(...,ret_B);
	     bvh.add_node_end(t,hit_B);
	*/
	vec4 add_node_start(vec4 &tex,const vec4 &nodebound) {
		if (tex.x>=tex.w) {
			printf("BVH: Added too many nodes!  x=%d, should be <w=%d\n",
				(int)tex.x, (int)tex.w);
			exit(1);
		}
		at(tex)=nodebound;
		tex.x+=1.0;
		if (tex.x<tex.w) // child's back link is our sibling, tex
			return tex;
		else // tex.x==tex.w: child's back link is our own back link
			return at(tex);
	}
	// Finish adding node, by linking in hit location.
	void add_node_end(vec4 tex,const vec4 &hitlink) {
		tex.x+=tex.z; // skip over into hit section
		at(tex)=hitlink;
	}
	
	/* Add a block of leaf objects, after which we render this next link. 
	   Returns destination tex coords for the new block. */
	vec4 add_leaf_block(int n,const vec4 &nextlink) {
		if (next_leaf>=h) { // leaf wraparound:
			leaf_x+=n+1; next_leaf=0;
		}
		vec4 tex(leaf_x,next_leaf++,0.0,leaf_x+n);
		at(tex+vec4(n,0.0,0.0,0.0))=nextlink; // add next link
		return tex;
	}
	
	/* Add another leaf object at tex.  Increments tex. */
	void add_leaf(vec4 &tex,const vec4 &leafdata) {
		if (tex.x>=tex.w) {
			printf("BVH: Added too many leaves!  x=%d, should be <w=%d\n",
				(int)tex.x, (int)tex.w);
			exit(1);
		}
		at(tex)=leafdata;
		tex.x+=1.0;
	}
	
	/* Upload texture data to GPU */
	void upload(GLhandleARB prog) {
		glFastUniform4fv(prog,"bvh_tex_root",1,root);

	/* Upload BVH texture, to texture unit 0 */
		glActiveTexture(GL_TEXTURE0);
		glTexImage2D(GL_TEXTURE_RECTANGLE_ARB,0,GL_RGBA32F_ARB, // internal format
			w,h,0, // width x height
			GL_RGBA,GL_FLOAT,&arr[0]);
		glTexParameteri(GL_TEXTURE_RECTANGLE_ARB,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
		glFastUniform1i(prog,"bvh_tex",0); // texture unit 0
	}
};

/* Create a bounding sphere from this matrix.  Maps unit sphere through matrix. */
vec4 make_sphere(const mat4 &mat) {
	vec4 center=vec4(0.0,0.0,0.0,1.0);
	vec4 radius=vec4(1.0,0.0,0.0,1.0);
	center=mat*center;
	radius=mat*radius;
	float r=length(radius-center);
	return vec4(center.x,center.y,center.z,r);
}

/* Rotation angle for this IFS */
float ifs_angle=0.0;
mat4 *ifs_mats=0;
void make_mats(void) {
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	delete[] ifs_mats;
	ifs_mats=new mat4[4]; /* initialize matrices */
	
	for (int n=0;n<4;n++) {
		glLoadIdentity();
		float l=0.5; // controls level-to-level scale factor
		float s=0.56*l; // controls spacing between spheres
		switch (n) {
		case 0: glTranslatef( s, s, s); glRotatef(ifs_angle, s, s, s); break;
		case 1: glTranslatef(-s,-s, s); glRotatef(ifs_angle,-s,-s, s); break;
		case 2: glTranslatef(-s, s,-s); glRotatef(ifs_angle,-s, s,-s); break;
		case 3: glTranslatef( s,-s,-s); glRotatef(ifs_angle, s,-s,-s); break;
		};
		glScalef(l,l,l);
		glGetFloatv(GL_MODELVIEW_MATRIX,&ifs_mats[n][0][0]);
	}
	glPopMatrix();
}

/* Return this IFS map */
mat4 make_mat(int number) {
	return ifs_mats[number];
}

/**
  Link up a recursive iterated function system bounding volume hierarchy.
  Returns the texture coordinates of the root BVH block.
*/
vec4 recursive_IFS(bvh_tex &bvh,const vec4 &backlink,const mat4 &mat,int level)
{
	int nkids=4;
	if (level<=0) 
	{ /* hit base level: add a set of leaf spheres */
		vec4 tex=bvh.add_leaf_block(nkids,backlink);
		vec4 ret=tex;
		for (int kid=0;kid<nkids;kid++) {
			bvh.add_leaf(tex,make_sphere(mat*make_mat(kid)));
		}
		return ret;
	}
	else /* interior level: add a node */
	{
		vec4 tex=bvh.add_node_block(nkids,backlink);
		vec4 ret=tex;
		for (int kid=0;kid<nkids;kid++) {
			mat4 kidmat=mat*make_mat(kid);
			vec4 nextlink=bvh.add_node_start(tex,make_sphere(kidmat));
			vec4 kidlink=recursive_IFS(bvh,nextlink,kidmat,level-1);
			bvh.add_node_end(tex,kidlink);
		}
		return ret;
	}
}


void display(void) 
{
	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE); // turn on Z writes
	glEnable(GL_BLEND);
	
	glClearColor(0.4,0.5,0.7,0.0); // background color
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); // flush any ancient matrices
	gluPerspective(70.0, // fov
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT),
		0.1,
		2000.0); // z clipping planes
	oglCameraLookAt(1.0);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity(); // flush any ancient matrices
	
	/* The proxy sphere should be gently rotating */
	const float rotate_rate=0.01; /* revolutions per second */
	float angle=glutGet(GLUT_ELAPSED_TIME)*0.001*rotate_rate*360;
	
	/* Build the pixel shader */
	static GLhandleARB prog=makeProgramObjectFromFiles(
		"raytrace_vertex.txt","raytrace.txt");
	glUseProgramObjectARB(prog);
	glFastUniform3fv(prog,"C",1,camera);
	
	//static bool uploadedtex=false;
	//if (!uploadedtex) {
	//	uploadedtex=true;
		// BVH texture data: RGBA width x height
		int level=4; // <- recursive levels of IFS expansion
		ifs_angle=angle; make_mats();
		bvh_tex bvh(9+4*(4+1),pow(4,level));
		bvh.root=recursive_IFS(bvh,bvh.root,mat4(1.0),level);
		bvh.upload(prog);
	//}
	
/* Upload checkerboard texture, to texture unit 1 */
	static bool hasChecker=false;
	if (!hasChecker) { /* only need to build and upload one time; it stays set up. */
		hasChecker=true;
		glActiveTexture(GL_TEXTURE1);
		int checkw=128,checkh=128;
		float *checkarr=new float[checkw*checkh];
		for (int y=0;y<checkh;y++) 
		for (int x=0;x<checkw;x++) {
			float cv=0.1;
			if ((x<checkw/2) ^ (y<checkh/2)) cv=0.9;
			checkarr[x+y*checkw]=cv;
		}
		gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGBA8,
			checkw,checkh,
			GL_LUMINANCE,GL_FLOAT,checkarr);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
		delete[] checkarr;
	}
	glFastUniform1i(prog,"checkertex",1); // texture unit 1
	
	
	// Only draw one set of faces, to avoid drawing raytraced geometry twice.
	//   (front side may get clipped if you're inside the object, so draw back)
	glEnable(GL_CULL_FACE); glCullFace(GL_FRONT);
	glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_ALPHA); // premultiplied alpha
	
	// Draw one *HUGE* sphere; this one proxy geometry covers all our objects.
	//   Keep in mind the far clipping plane in determining the sphere radius
	glutSolidSphere(1000.0,3,3); 
	
	glUseProgramObjectARB(0);
	
	glutPostRedisplay(); // continual animation
	glutSwapBuffers();

/* Estimate framerate */
	static double last_time=0.0;
	static int framecount=0, last_framecount=0;
	framecount++;
	double cur_time=0.001*glutGet(GLUT_ELAPSED_TIME);
	if (cur_time>0.5+last_time) 
	{ // every half a second, estimate fps
		double time=cur_time-last_time;
		double time_per_frame=time/(framecount-last_framecount);
		char str[100];
		sprintf(str,"IFS Raytracer: %.1f fps, %.1f ms/frame",
			1.0/time_per_frame,1.0e3*time_per_frame);
		glutSetWindowTitle(str);
		last_framecount=framecount;
		last_time=cur_time;
	}
}

int main(int argc,char *argv[]) 
{
	glutInit(&argc,argv);
	glutInitDisplayMode(GLUT_RGBA + GLUT_DEPTH + GLUT_DOUBLE);
	glutInitWindowSize(800,600);
	glutCreateWindow("GPU Raytracer");
	
	glewInit();
	
	glutDisplayFunc(display);
	oglCameraInit();
	camera=vec3(0,-2,0.3);
	
	glutMainLoop();
	return 0;
}
